#-------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/18/2005
#
#-------------------------------------------------------------------------------

""" Pyface 'DockWindow' support.

    This package provides a Pyface 'dockable' window component that allows
    child windows to be reorganized within the DockWindow using drag and drop.
    The component also allows multiple sub-windows to occupy the same
    sub-region of the DockWindow, in which case each sub-window appears as a
    separate notebook-like tab within the region.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import shelve
import os
import wx
import sys

from pyface.api import SystemMetrics

from traits.api \
    import HasPrivateTraits, Instance, Tuple, Property, Any, Str, List, false

from traits.trait_base \
    import traits_home

from traitsui.api \
    import View, HGroup, VGroup, Item, Handler, error

from traitsui.helper \
    import user_name_for

from traitsui.menu \
    import Menu, Action, Separator

from traitsui.dockable_view_element \
    import DockableViewElement

from traitsui.dock_window_theme \
    import dock_window_theme, DockWindowTheme

from pyface.wx.drag_and_drop \
    import PythonDropTarget, clipboard

from pyface.message_dialog \
    import error as warning

from .dock_sizer \
    import DockSizer, DockControl, DockRegion, DockStyle, DockSplitter, \
           no_dock_info, clear_window, features

from .idockable \
    import IDockable

from .idock_ui_provider \
    import IDockUIProvider

is_mac = (sys.platform == 'darwin')

#-------------------------------------------------------------------------------
#  Global data:
#-------------------------------------------------------------------------------

# Dictionary of cursors in use:
cursor_map = {}

#-------------------------------------------------------------------------------
#  DockWindow context menu:
#-------------------------------------------------------------------------------

min_max_action = Action( name   = 'Maximize',
                         action = 'on_min_max' )

undock_action  = Action( name   = 'Undock',
                         action = 'on_undock' )

lock_action    = Action( name   = 'Lock Layout',
                         action = 'on_lock_layout',
                         style  = 'toggle' )

layout_action  = Action( name   = 'Switch Layout',
                         action = 'on_switch_layout' )

save_action    = Action( name   = 'Save Layout...',
                         action = 'on_save_layout' )

#hide_action    = Action( name   = 'Hide',
#                         action = 'on_hide' )
#
#show_action    = Action( name   = 'Show',
#                         action = 'on_show' )

edit_action    = Action( name   = 'Edit Properties...',
                         action = 'on_edit' )

enable_features_action  = Action( name   = 'Show All',
                                  action = 'on_enable_all_features' )

disable_features_action = Action( name   = 'Hide All',
                                  action = 'on_disable_all_features' )

#-------------------------------------------------------------------------------
#  'DockWindowHandler' class/interface:
#-------------------------------------------------------------------------------

class DockWindowHandler ( HasPrivateTraits ):

    #---------------------------------------------------------------------------
    #  Returns whether or not a specified object can be inserted into the view:
    #---------------------------------------------------------------------------

    def can_drop ( self, object ):
        """ Returns whether or not a specified object can be inserted into the
            view.
        """
        return True

    #---------------------------------------------------------------------------
    #  Returns the DockControl object for a specified object:
    #---------------------------------------------------------------------------

    def dock_control_for ( self, parent, object ):
        """ Returns the DockControl object for a specified object.
        """
        try:
            name = object.name
        except:
            try:
                name = object.label
            except:
                name = ''
        if name == '':
            name = user_name_for( object.__class__.__name__ )

        image  = None
        export = ''
        if isinstance( object, DockControl ):
            dock_control = object
            image        = dock_control.image
            export       = dock_control.export
            dockable     = dock_control.dockable
            close        = dockable.dockable_should_close()
            if close:
                dock_control.close( force = True )

            control = dockable.dockable_get_control( parent )

            # If DockControl was closed, then reset it to point to the new
            # control:
            if close:
                dock_control.set( control = control,
                                  style   = parent.owner.style )
                dockable.dockable_init_dockcontrol( dock_control )
                return dock_control

        elif isinstance( object, IDockable ):
            dockable = object
            control  = dockable.dockable_get_control( parent )
        else:
            ui       = object.get_dockable_ui( parent )
            dockable = DockableViewElement( ui = ui )
            export   = ui.view.export
            control  = ui.control

        dc = DockControl( control   = control,
                          name      = name,
                          export    = export,
                          style     = parent.owner.style,
                          image     = image,
                          closeable = True )

        dockable.dockable_init_dockcontrol( dc )

        return dc

    #---------------------------------------------------------------------------
    #  Creates a new view of a specified control:
    #---------------------------------------------------------------------------

    def open_view_for ( self, control, use_mouse = True ):
        """ Creates a new view of a specified control.
        """
        from .dock_window_shell import DockWindowShell

        DockWindowShell( control, use_mouse = use_mouse )

    #---------------------------------------------------------------------------
    #  Handles the DockWindow becoming empty:
    #---------------------------------------------------------------------------

    def dock_window_empty ( self, dock_window ):
        """ Handles the DockWindow becoming empty.
        """
        if dock_window.auto_close:
            dock_window.control.GetParent().Destroy()

# Create a singleton handler:
dock_window_handler = DockWindowHandler()

#-------------------------------------------------------------------------------
#  'LayoutName' class:
#-------------------------------------------------------------------------------

class LayoutName ( Handler ):

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Name the user wants to assign to the layout:
    name = Str

    # List of currently assigned names:
    names = List( Str )

    #---------------------------------------------------------------------------
    #  Traits view definitions:
    #---------------------------------------------------------------------------

    view = View( Item( 'name', label = 'Layout name' ),
                 title   = 'Save Layout',
                 kind    = 'modal',
                 buttons = [ 'OK', 'Cancel' ] )

    #---------------------------------------------------------------------------
    #  Handles a request to close a dialog-based user interface by the user:
    #---------------------------------------------------------------------------

    def close ( self, info, is_ok ):
        if is_ok:
            name = info.object.name.strip()
            if name == '':
                warning( info.ui.control, 'No name specified',
                         title = 'Save Layout Error' )
                return False
            if name in self.names:
                return error( message = '%s is already defined. Replace?' %
                                        name,
                              title   = 'Save Layout Warning',
                              parent  = info.ui.control )

        return True

#-------------------------------------------------------------------------------
#  'DockWindow' class:
#-------------------------------------------------------------------------------

class DockWindow ( HasPrivateTraits ):

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # The wx.Window being used as the DockWindow:
    control = Instance( wx.Window )

    # The handler used to determine how certain events should be processed:
    handler = Any( dock_window_handler )

    # The 'extra' arguments to be passed to each handler call:
    handler_args = Tuple

    # Close the parent window if the DockWindow becomes empty:
    auto_close = false

    # The DockWindow graphical theme style information:
    theme = Instance( DockWindowTheme, allow_none = False )

    # Default style for external objects dragged into the window:
    style = DockStyle

    # Return the sizer associated with the control (i.e. window)
    sizer = Property

    # The id used to identify this DockWindow:
    id = Str

    #---------------------------------------------------------------------------
    #  Initializes the object:
    #---------------------------------------------------------------------------

    def __init__ ( self, parent, wid = -1, pos = wx.DefaultPosition,
                   size = wx.DefaultSize, style = wx.FULL_REPAINT_ON_RESIZE,
                   **traits ):
        super( DockWindow, self ).__init__( **traits )

        # Create the actual window:
        self.control  = control = wx.Window( parent, wid, pos, size, style )
        control.owner = self

        # Set up the 'paint' event handler:
        wx.EVT_PAINT( control, self._paint )
        wx.EVT_SIZE(  control, self._size )

        # Set up mouse event handlers:
        wx.EVT_LEFT_DOWN(    control, self._left_down )
        wx.EVT_LEFT_UP(      control, self._left_up )
        wx.EVT_LEFT_DCLICK(  control, self._left_dclick )
        wx.EVT_RIGHT_DOWN(   control, self._right_down )
        wx.EVT_RIGHT_UP(     control, self.right_up )
        wx.EVT_MOTION(       control, self._mouse_move )
        wx.EVT_LEAVE_WINDOW( control, self._mouse_leave )

        control.SetDropTarget( PythonDropTarget( self ) )

        # Initialize the window background color:
        if self.theme.use_theme_color:
            color = self.theme.tab.image_slice.bg_color
        else:
            color = SystemMetrics().dialog_background_color
            color = wx.Colour(color[0]*255, color[1]*255, color[2]*255)

        self.control.SetBackgroundColour( color )

    #-- Default Trait Values ---------------------------------------------------

    def _theme_default ( self ):
        return dock_window_theme()

    #-- Trait Event Handlers ---------------------------------------------------

    def _theme_changed ( self, theme ):
        if self.control is not None:
            if theme.use_theme_color:
                color = theme.tab.image_slice.bg_color
            else:
                color = wx.NullColour
            self.control.SetBackgroundColour( color )

            self.update_layout()

    #---------------------------------------------------------------------------
    #  Notifies the DockWindow that its contents are empty:
    #---------------------------------------------------------------------------

    def dock_window_empty ( self ):
        """ Notifies the DockWindow that its contents are empty.
        """
        self.handler.dock_window_empty( self )

    #---------------------------------------------------------------------------
    #  Sets the cursor to a specified cursor shape:
    #---------------------------------------------------------------------------

    def set_cursor ( self, cursor = None ):
        """ Sets the cursor to a specified cursor shape.
        """
        if cursor is None:
            self.control.SetCursor( wx.NullCursor )
            return

        global cursor_map

        if cursor not in cursor_map:
            cursor_map[ cursor ] = wx.StockCursor( cursor )

        self.control.SetCursor( cursor_map[ cursor ] )

    #---------------------------------------------------------------------------
    #  Releases ownership of the mouse capture:
    #---------------------------------------------------------------------------

    def release_mouse ( self ):
        """ Releases ownership of the mouse capture.
        """
        if self._owner is not None:
            self._owner = None
            self.control.ReleaseMouse()

    #---------------------------------------------------------------------------
    #  Updates the layout of the window:
    #---------------------------------------------------------------------------

    def update_layout ( self ):
        """ Updates the layout of the window.
        """
        # There are cases where a layout has been scheduled for a DockWindow,
        # but then the DockWindow is destroyed, which will cause the calls
        # below to fail. So we catch the 'PyDeadObjectError' exception and
        # ignore it:
        try:
            self.control.Layout()
            self.control.Refresh()
        except wx.PyDeadObjectError:
            pass

    #---------------------------------------------------------------------------
    #  Minimizes/Maximizes a specified DockControl:
    #---------------------------------------------------------------------------

    def min_max ( self, dock_control ):
        """ Minimizes/maximizes a specified DockControl.
        """
        sizer = self.sizer
        if sizer is not None:
            sizer.MinMax( self.control, dock_control )
            self.update_layout()

    #---------------------------------------------------------------------------
    #  Pops up the feature bar for a specified DockControl:
    #---------------------------------------------------------------------------

    def feature_bar_popup ( self, dock_control ):
        """ Pops up the feature bar for a specified DockControl.
        """
        fb = self._feature_bar
        if fb is None:
            from .feature_bar import FeatureBar

            self._feature_bar = fb = FeatureBar( parent = self.control )
            fb.on_trait_change( self._feature_bar_closed, 'completed' )

        fb.dock_control = dock_control
        fb.show()

    #---------------------------------------------------------------------------
    #  Handles closing the feature bar:
    #---------------------------------------------------------------------------

    def _feature_bar_closed ( self ):
        fb = self._feature_bar
        fb.dock_control.feature_bar_closed()
        fb.hide()

    #---------------------------------------------------------------------------
    #  Perform all operations needed to close the window:
    #---------------------------------------------------------------------------

    def close ( self ):
        """ Closes the dock window.  In this case, all event handlers are
        unregistered.  Other cleanup operations go here, but at the moment Linux
        (and other non-Windows platforms?) are less forgiving when things like
        event handlers arent unregistered.
        """
        self._unregister_event_handlers()

    #---------------------------------------------------------------------------
    #  Unregister all event handlers:
    #---------------------------------------------------------------------------

    def _unregister_event_handlers ( self ):
        """ Unregister all event handlers setup in the constructor.  This is
        typically done prior to an app shutting down and is needed since Linux
        (and other non-Windows platforms?) trigger mouse, repaint, etc. events
        for controls which have already been deleted.
        """
        control = self.control
        if control is not None:
            wx.EVT_PAINT(            control, None )
            wx.EVT_SIZE(             control, None )
            wx.EVT_LEFT_DOWN(        control, None )
            wx.EVT_LEFT_UP(          control, None )
            wx.EVT_LEFT_DCLICK(      control, None )
            wx.EVT_RIGHT_DOWN(       control, None )
            wx.EVT_RIGHT_UP(         control, None )
            wx.EVT_MOTION(           control, None )
            wx.EVT_LEAVE_WINDOW(     control, None )

    #---------------------------------------------------------------------------
    #  Handles repainting the window:
    #---------------------------------------------------------------------------

    def _paint ( self, event ):
        """ Handles repainting the window.
        """
        # There is a problem on macs where we get paints when the update
        # is entirely within children.
        if is_mac and self._is_child_paint():
            return

        sizer = self.sizer
        if isinstance( sizer, DockSizer ):
            sizer.Draw( self.control )
        else:
            clear_window( self.control )

    #---------------------------------------------------------------------------
    #  Uses wx calls to determine if we really need to paint or the children will
    #  do it.
    #---------------------------------------------------------------------------

    def _is_child_paint ( self ):
        """ Returns whether or not the current update region is entirely within a child.
        """
        if self.control.Children:
            update_rect = self.control.UpdateRegion.Box
            for child in self.control.Children:
                if not child.HasTransparentBackground() and \
                        child.Rect.ContainsRect(update_rect):
                    return True
        return False

    #---------------------------------------------------------------------------
    #  Handles the window being resized:
    #---------------------------------------------------------------------------

    def _size ( self, event ):
        """ Handles the window being resized.
        """
        sizer = self.sizer
        if sizer is not None:
            try:
                dx, dy = self.control.GetSizeTuple()
                sizer.SetDimension( 0, 0, dx, dy )
            except:
                # fixme: This is temporary code to work around a problem in
                #        ProAVA2 that we are still trying to track down...
                pass

    #---------------------------------------------------------------------------
    #  Handles the left mouse button being pressed:
    #---------------------------------------------------------------------------

    def _left_down ( self, event ):
        """ Handles the left mouse button being pressed.
        """
        sizer = self.sizer
        if sizer is not None:
            object = sizer.ObjectAt( event.GetX(), event.GetY() )
            if object is not None:
                self._owner = object
                self.control.CaptureMouse()
                object.mouse_down( event )

    #---------------------------------------------------------------------------
    #  Handles the left mouse button being released:
    #---------------------------------------------------------------------------

    def _left_up ( self, event ):
        """ Handles the left mouse button being released.
        """
        window = self.control
        if self._owner is not None:
            window.ReleaseMouse()
            self._owner.mouse_up( event )
            self._owner = None

        # Check for the user requesting that the layout structure be reset:
        if event.ShiftDown():
            if event.ControlDown():
                # Check for the developer requesting a structure dump (DEBUG):
                if event.AltDown():
                    contents = self.sizer.GetContents()
                    if contents is not None:
                        contents.dump()
                        sys.stdout.flush()
                else:
                    self.sizer.ResetStructure( window )
                    self.update_layout()
            elif event.AltDown():
                self.sizer.ToggleLock()
                self.update_layout()

    #---------------------------------------------------------------------------
    #  Handles the left mouse button being double clicked:
    #---------------------------------------------------------------------------

    def _left_dclick ( self, event ):
        """ Handles the left mouse button being double clicked.
        """
        sizer = self.sizer
        if sizer is not None:
            object = sizer.ObjectAt( event.GetX(), event.GetY(), True )
            if isinstance( object, DockControl ):
                dockable = object.dockable
                if (((dockable is None) or
                     (dockable.dockable_dclick( object, event ) is False)) and
                    (object.style != 'fixed')):
                    self.min_max( object )
            elif isinstance( object, DockRegion ):
                self._owner = object
                self.control.CaptureMouse()
                object.mouse_down( event )

    #---------------------------------------------------------------------------
    #  Handles the right mouse button being pressed:
    #---------------------------------------------------------------------------

    def _right_down ( self, event ):
        """ Handles the right mouse button being pressed.
        """
        pass

    #---------------------------------------------------------------------------
    #  Handles the right mouse button being released:
    #---------------------------------------------------------------------------

    def right_up ( self, event ):
        """ Handles the right mouse button being released.
        """
        sizer = self.sizer
        if sizer is not None:
            object = sizer.ObjectAt( event.GetX(), event.GetY(), True )
            if object is not None:
                popup_menu      = None
                window          = self.control
                is_dock_control = isinstance( object, DockControl )

                if (is_dock_control and (object.dockable is not None) and
                    (event.ShiftDown() or event.ControlDown() or
                     event.AltDown())):
                    self._menu_self = object.dockable
                    popup_menu = object.dockable.dockable_menu( object, event )

                if popup_menu is None:
                    self._menu_self = self
                    section         = self.sizer.GetContents()
                    is_splitter     = isinstance( object, DockSplitter )
                    self._object    = object
                    if is_splitter:
                        self._object = object = object.parent
                    group = object
                    if is_dock_control:
                        group = group.parent
                    if sizer.IsMaximizable():
                        min_max_action.name = 'Maximize'
                    else:
                        min_max_action.name = 'Restore'
                    min_max_action.enabled  = is_dock_control
                    undock_action.enabled   = is_dock_control
                    edit_action.enabled     = (not is_splitter)
                    controls                = section.get_controls( False )
                    lock_action.checked     = ((len( controls ) > 0) and
                                               controls[0].locked)
                    save_action.enabled     = (self.id != '')

                    feature_menu = self._get_feature_menu()
                    restore_menu, delete_menu = self._get_layout_menus()
                    popup_menu = Menu( min_max_action,
                                       undock_action,
                                       Separator(),
                                       feature_menu,
                                       #Separator(),
                                       #hide_action,
                                       #show_action,
                                       Separator(),
                                       lock_action,
                                       layout_action,
                                       Separator(),
                                       save_action,
                                       restore_menu,
                                       delete_menu,
                                       Separator(),
                                       edit_action,
                                       name = 'popup' )

                window.PopupMenuXY( popup_menu.create_menu( window, self ),
                                    event.GetX() - 10, event.GetY() - 10 )
                self._object = None

    #---------------------------------------------------------------------------
    #  Handles the mouse moving over the window:
    #---------------------------------------------------------------------------

    def _mouse_move ( self, event ):
        """ Handles the mouse moving over the window.
        """
        if self._last_dock_control is not None:
            self._last_dock_control.reset_feature_popup()
            self._last_dock_control = None

        if self._owner is not None:
            self._owner.mouse_move( event )
        else:
            sizer = self.sizer
            if sizer is not None:
                object = (self._object or
                          sizer.ObjectAt( event.GetX(), event.GetY() ))
                self._set_cursor( event, object )

                if object is not self._hover:
                    if self._hover is not None:
                        self._hover.hover_exit( event )

                    if object is not None:
                        object.hover_enter( event )

                    self._hover = object

                # Handle feature related processing:
                if (isinstance( object, DockControl ) and
                    object.feature_activate( event )):
                    self._last_dock_control = object


    #---------------------------------------------------------------------------
    #  Handles the mouse leaving the window:
    #---------------------------------------------------------------------------

    def _mouse_leave ( self, event ):
        """ Handles the mouse leaving the window.
        """
        if self._hover is not None:
            self._hover.hover_exit( event )
            self._hover = None
        self._set_cursor( event )

    #---------------------------------------------------------------------------
    #  Sets the cursor for a specified object:
    #---------------------------------------------------------------------------

    def _set_cursor ( self, event, object = None ):
        """ Sets the cursor for a specified object.
        """
        if object is None:
            self.set_cursor()
        else:
            self.set_cursor( object.get_cursor( event ) )

#-- Context menu action handlers -----------------------------------------------

    #---------------------------------------------------------------------------
    #  Handles the user asking for a DockControl to be maximized/restored:
    #---------------------------------------------------------------------------

    def on_min_max ( self ):
        """ Handles the user asking for a DockControl to be maximized/restored.
        """
        self.min_max( self._object )

    #---------------------------------------------------------------------------
    #  Handles the user requesting an element to be undocked:
    #---------------------------------------------------------------------------

    def on_undock ( self ):
        """ Handles the user requesting an element to be undocked.
        """
        self.handler.open_view_for( self._object, use_mouse = False )

    #---------------------------------------------------------------------------
    #  Handles the user requesting an element to be hidden:
    #---------------------------------------------------------------------------

    def on_hide ( self ):
        """ Handles the user requesting an element to be hidden.
        """
        self._object.show( False )

    #---------------------------------------------------------------------------
    #  Handles the user requesting an element to be shown:
    #---------------------------------------------------------------------------

    def on_show ( self ):
        """ Handles the user requesting an element to be shown.
        """
        object = self._object
        if isinstance( object, DockControl ):
            object = object.parent
        self._hidden_group_for( object ).show( True )

    #---------------------------------------------------------------------------
    #  Handles the user requesting that the current layout be switched:
    #---------------------------------------------------------------------------

    def on_switch_layout ( self ):
        """ Handles the user requesting that the current layout be switched.
        """
        self.sizer.ResetStructure( self.control )
        self.update_layout()

    #---------------------------------------------------------------------------
    #  Handles the user requesting that the layout be locked/unlocked:
    #---------------------------------------------------------------------------

    def on_lock_layout ( self ):
        """ Handles the user requesting that the layout be locked/unlocked.
        """
        self.sizer.ToggleLock()
        self.update_layout()

    #---------------------------------------------------------------------------
    #  Handles the user requesting that the layout be saved:
    #---------------------------------------------------------------------------

    def on_save_layout ( self ):
        """ Handles the user requesting that the layout be saved.
        """
        layout_name = LayoutName( names = self._get_layout_names() )
        if layout_name.edit_traits( parent = self.control ).result:
            self._set_layout( layout_name.name, self.sizer.GetStructure() )

    #---------------------------------------------------------------------------
    #  Handles the user requesting a specified layout to be restored:
    #---------------------------------------------------------------------------

    def on_restore_layout ( self, name ):
        """ Handles the user requesting a specified layout to be restored.
        """
        self.sizer.SetStructure( self.control, self._get_layout( name ) )
        self.update_layout()

    #---------------------------------------------------------------------------
    #  Handles the user reqesting a specified layout to be deleted:
    #---------------------------------------------------------------------------

    def on_delete_layout ( self, name ):
        """ Handles the user reqesting a specified layout to be deleted.
        """
        if error( message = "Delete the '%s' layout?" % name,
                  title   = 'Delete Layout Warning',
                  parent  = self.control ):
            self._delete_layout( name )

    #---------------------------------------------------------------------------
    #  Handles the user requesting to edit an item:
    #---------------------------------------------------------------------------

    def on_edit ( self, object = None ):
        """ Handles the user requesting to edit an item.
        """
        if object is None:
            object = self._object
        control_info = ControlInfo( **object.get( 'name', 'user_name',
                                                  'style', 'user_style' ) )
        if control_info.edit_traits( parent = self.control ).result:
            name = control_info.name.strip()
            if name != '':
                object.name = name
            object.set( **control_info.get( 'user_name',
                                            'style', 'user_style' ) )
            self.update_layout()

    #---------------------------------------------------------------------------
    #  Enables all features:
    #---------------------------------------------------------------------------

    def on_enable_all_features ( self, action ):
        """ Enables all features.
        """
        for feature in features:
            if (feature.feature_name != '') and (feature.state != 1):
                feature.toggle_feature( action )

    #---------------------------------------------------------------------------
    #  Disables all features:
    #---------------------------------------------------------------------------

    def on_disable_all_features ( self, action ):
        """ Disables all features.
        """
        for feature in features:
            if (feature.feature_name != '') and (feature.state == 1):
                feature.toggle_feature( action )

    #---------------------------------------------------------------------------
    #  Toggles the enabled/disabled state of the action's associated feature:
    #---------------------------------------------------------------------------

    def on_toggle_feature ( self, action ):
        """ Toggles the enabled/disabled state of the action's associated
            feature.
        """
        action._feature.toggle_feature( action )

#-- DockWindow user preference database methods --------------------------------

    #---------------------------------------------------------------------------
    #  Gets the layout dictionary for the DockWindow:
    #---------------------------------------------------------------------------

    def _get_layouts ( self ):
        """ Gets the layout dictionary for the DockWindow.
        """
        id = self.id
        if id != '':
            db = self._get_dw_db()
            if db is not None:
                layouts = db.get( id )
                db.close()
                return layouts

        return None

    #---------------------------------------------------------------------------
    #  Gets the names of all current layouts defined for the DockWindow:
    #---------------------------------------------------------------------------

    def _get_layout_names ( self ):
        """ Gets the names of all current layouts defined for the DockWindow.
        """
        layouts = self._get_layouts()
        if layouts is not None:
            return list(layouts.keys())

        return []

    #---------------------------------------------------------------------------
    #  Gets the layout data for a specified layout name:
    #---------------------------------------------------------------------------

    def _get_layout ( self, name ):
        """ Gets the layout data for a specified layout name.
        """
        layouts = self._get_layouts()
        if layouts is not None:
            return layouts.get( name )

        return None

    #---------------------------------------------------------------------------
    #  Deletes the layout data for a specified layout name:
    #---------------------------------------------------------------------------

    def _delete_layout ( self, name ):
        """ Deletes the layout data for a specified layout name.
        """
        id = self.id
        if id != '':
            db = self._get_dw_db( mode = 'c' )
            if db is not None:
                layouts = db.get( id )
                if layouts is not None:
                    del layouts[ name ]
                    db[ id ] = layouts
                db.close()

    #---------------------------------------------------------------------------
    #  Sets the layout data for a specified layout name:
    #---------------------------------------------------------------------------

    def _set_layout ( self, name, layout ):
        """ Sets the layout data for a specified layout name.
        """
        id = self.id
        if id != '':
            db = self._get_dw_db( mode = 'c' )
            if db is not None:
                layouts = db.get( id )
                if layouts is None:
                    layouts = {}
                layouts[ name ] = layout
                db[ id ] = layouts
                db.close()

    #---------------------------------------------------------------------------
    #  Gets a reference to the DockWindow UI preference database:
    #---------------------------------------------------------------------------

    def _get_dw_db ( self, mode = 'r' ):
        try:
            return shelve.open( os.path.join( traits_home(), 'dock_window' ),
                                flag = mode, protocol = -1 )
        except:
            return None

    #---------------------------------------------------------------------------
    #  Returns the 'Features' sub_menu:
    #---------------------------------------------------------------------------

    def _get_feature_menu ( self ):
        """ Returns the 'Features' sub_menu.
        """
        enable_features_action.enabled = disable_features_action.enabled = False
        for feature in features:
            if feature.feature_name != '':
                if feature.state == 1:
                    disable_features_action.enabled = True
                    if enable_features_action.enabled:
                        break
                else:
                    enable_features_action.enabled = True
                    if disable_features_action.enabled:
                        break

        actions = []
        for feature in features:
            if feature.feature_name != '':
                actions.append( Action( name     = feature.feature_name,
                                        action   = 'on_toggle_feature',
                                        _feature = feature,
                                        style    = 'toggle',
                                        checked  = (feature.state == 1) ) )

        if len( actions ) > 0:
            actions.sort( lambda l, r: cmp( l.name, r.name ) )
            actions[0:0] = [ Separator() ]

        return Menu( name = 'Features',
                     *([ enable_features_action, disable_features_action ] +
                       actions) )

    #---------------------------------------------------------------------------
    #  Gets the sub-menus for the 'Restore layout' and 'Delete layout' menu
    #  options:
    #---------------------------------------------------------------------------

    def _get_layout_menus ( self ):
        """ Gets the sub-menus for the 'Restore layout' and 'Delete layout' menu
            options.
        """
        names = self._get_layout_names()
        if len( names ) == 0:
            restore_actions = [ Action( name = '<empty>', enabled = False ) ]
            delete_actions  = [ Action( name = '<empty>', enabled = False ) ]
        else:
            names.sort()
            restore_actions = [ Action(
                                    name   = name,
                                    action = "self.on_restore_layout(%s)" %
                                             repr( name ) )
                                for name in names ]
            delete_actions  = [ Action(
                                    name   = name,
                                    action = "self.on_delete_layout(%s)" %
                                             repr( name ) )
                                for name in names ]
        return [ Menu( name = 'Restore Layout', *restore_actions ),
                 Menu( name = 'Delete Layout',  *delete_actions ) ]

#-- Drag and drop event handlers: ----------------------------------------------

    #---------------------------------------------------------------------------
    #  Handles a Python object being dropped on the control:
    #---------------------------------------------------------------------------

    def wx_dropped_on ( self, x, y, data, drag_result ):
        """ Handles a Python object being dropped on the window.
        """
        if isinstance( data, ( IDockUIProvider, DockControl ) ):
            window    = self.control
            dock_info = self._dock_info

            # See the 'wx_drag_leave' method for an explanation of this code:
            if dock_info is None:
                dock_info = self._leave_info

            dock_info.draw( window )
            self._dock_info = None
            try:
                control = self.handler.dock_control_for(
                                       *(self.handler_args + ( window, data )) )
                # Safely check to see if the object quacks like a Binding
                binding = getattr( clipboard, 'node', None )
                if (hasattr(binding, "obj") and (binding.obj is data) and
                        hasattr(binding, "namespace_name")):
                    control.id = '@@%s' % binding.namespace_name
                dock_info.dock( control, window )
                return drag_result
            except:
                warning( window,
                         "An error occurred while attempting to add an item of "
                         "type '%s' to the window." % data.__class__.__name__,
                         title = 'Cannot add item to window' )

        return wx.DragNone

    #---------------------------------------------------------------------------
    #  Handles a Python object being dragged over the control:
    #---------------------------------------------------------------------------

    def wx_drag_any ( self, x, y, data, drag_result ):
        """ Handles a Python object being dragged over the control.
        """
        ui_provider = isinstance( data, IDockUIProvider )
        if ui_provider or isinstance( data, DockControl ):
            if (ui_provider and
               (not self.handler.can_drop( *(self.handler_args + ( data, )) ))):
                return wx.DragNone

            # Check to see if we are in 'drag mode' yet:
            cur_dock_info = self._dock_info
            if cur_dock_info is None:
                cur_dock_info = no_dock_info
                if isinstance( data, DockControl ):
                    self._dock_size = data.tab_width
                else:
                    self._dock_size = 80

            # Get the window and DockInfo object associated with the event:
            window          = self.control
            self._dock_info = dock_info = \
                              self.sizer.DockInfoAt( x, y, self._dock_size,
                                                     False )

            # If the DockInfo has changed, then update the screen:
            if ((cur_dock_info.kind != dock_info.kind)         or
                (cur_dock_info.region is not dock_info.region) or
                (cur_dock_info.bounds != dock_info.bounds)     or
                (cur_dock_info.tab_bounds != dock_info.tab_bounds)):

                # Erase the old region:
                cur_dock_info.draw( window )

                # Draw the new region:
                dock_info.draw( window )

            return drag_result

        # Handle the case of dragging a normal object over a 'feature':
        if self._can_drop_on_feature( x, y, data ) is not None:
            return drag_result

        return wx.DragNone

    #---------------------------------------------------------------------------
    #  Handles a dragged Python object leaving the window:
    #---------------------------------------------------------------------------

    def wx_drag_leave ( self, data ):
        """ Handles a dragged Python object leaving the window.
        """
        if self._dock_info is not None:
            self._dock_info.draw( self.control )

            # Save the current '_dock_info' in '_leave_info' because under
            # Linux the drag and drop code sends a spurious 'drag_leave' event
            # immediately before a 'dropped_on' event, so we need to remember
            # the '_dock_info' value just in case the next event is
            # 'dropped_on':
            self._leave_info, self._dock_info = self._dock_info, None

#-- Pyface menu interface implementation ---------------------------------------

    #---------------------------------------------------------------------------
    #  Adds a menu item to the menu bar being constructed:
    #---------------------------------------------------------------------------

    def add_to_menu ( self, menu_item ):
        """ Adds a menu item to the menu bar being constructed.
        """
        pass

    #---------------------------------------------------------------------------
    #  Adds a tool bar item to the tool bar being constructed:
    #---------------------------------------------------------------------------

    def add_to_toolbar ( self, toolbar_item ):
        """ Adds a tool bar item to the tool bar being constructed.
        """
        pass

    #---------------------------------------------------------------------------
    #  Returns whether the menu action should be defined in the user interface:
    #---------------------------------------------------------------------------

    def can_add_to_menu ( self, action ):
        """ Returns whether the action should be defined in the user interface.
        """
        return True

    #---------------------------------------------------------------------------
    #  Returns whether the toolbar action should be defined in the user
    #  interface:
    #---------------------------------------------------------------------------

    def can_add_to_toolbar ( self, action ):
        """ Returns whether the toolbar action should be defined in the user
            interface.
        """
        return True

    #---------------------------------------------------------------------------
    #  Performs the action described by a specified Action object:
    #---------------------------------------------------------------------------

    def perform ( self, action_object ):
        """ Performs the action described by a specified Action object.
        """
        action = action_object.action
        if action[ : 5 ] == 'self.':
            eval( action, globals(), { 'self': self._menu_self } )
        else:
            method = getattr( self, action )
            try:
                method()
            except:
                method( action_object )

#-- Property implementations ---------------------------------------------------

    #---------------------------------------------------------------------------
    #  Implementation of the 'sizer' property:
    #---------------------------------------------------------------------------

    def _get_sizer ( self ):
        if self.control is not None:
            return self.control.GetSizer()
        return None

    def _set_sizer ( self, sizer ):
        self.control.SetSizer( sizer )

#-- Private methods ------------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Finds the first group with any hidden items (if any):
    #---------------------------------------------------------------------------

    def _hidden_group_for ( self, group ):
        """ Finds the first group with any hidden items (if any).
        """
        while True:
            if group is None:
                return None
            if len( group.contents ) > len( group.visible_contents ):
                return group
            group = group.parent

    #---------------------------------------------------------------------------
    #  Returns a feature that the pointer is over and which can accept the
    #  specified data:
    #---------------------------------------------------------------------------

    def _can_drop_on_feature ( self, x, y, data ):
        """ Returns a feature that the pointer is over and which can accept the
            specified data.
        """
        if self.sizer is not None:
            object = self.sizer.ObjectAt( x, y )
            if isinstance( object, DockControl ):
                event = FakeEvent( x, y, self.control )

                if object.feature_activate( event, data ):
                    ldc = self._last_dock_control
                    if (ldc is not None) and (ldc is not object):
                        ldc.reset_feature_popup()
                    self._last_dock_control = object
                    return None

        if self._last_dock_control is not None:
            self._last_dock_control.reset_feature_popup()
            self._last_dock_control = None

        return None

#-------------------------------------------------------------------------------
#  'FakeEvent' class:
#-------------------------------------------------------------------------------

class FakeEvent ( object ):

    def __init__ ( self, x, y, object ):
        self.x, self.y, self.object = x, y, object

    def GetX ( self ): return self.x
    def GetY ( self ): return self.y
    def GetEventObject ( self ): return self.object

#-------------------------------------------------------------------------------
#  'ControlInfo' class:
#-------------------------------------------------------------------------------

class ControlInfo ( HasPrivateTraits ):

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Name to be edited:
    name = Str

    # Has the user set the name of the control?
    user_name = false
    id = Str

    # Style of drag bar/tab:
    style = DockStyle

    # Has the user set the style for this control:
    user_style = false

    #---------------------------------------------------------------------------
    #  Traits view definition:
    #---------------------------------------------------------------------------

    traits_view = View( VGroup(
                            HGroup( HGroup( 'name<100>{Label}', '3' ),
                                    HGroup( 'user_name{Remember label}',
                                            show_left = False ) ),
                            HGroup( HGroup( 'style<101>', '3' ),
                                    HGroup( 'user_style{Remember style}',
                                            show_left = False ) ) ),
                        title   = 'Edit properties',
                        kind    = 'modal',
                        buttons = [ 'OK', 'Cancel' ] )




#------------------------------------------------------------------------------
# Copyright (c) 2009, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Evan Patterson
# Date: 06/26/09
#------------------------------------------------------------------------------

# Standard library imports
from io import BytesIO
from pickle import dumps, load, loads

# System library imports
from pyface.qt import QtCore, QtGui

# ETS imports
from traits.api import provides
from pyface.i_clipboard import IClipboard, BaseClipboard

# Shortcuts
cb = QtGui.QApplication.clipboard()

# Custom MIME type representing python objects
PYTHON_TYPE = "python/object"


@provides(IClipboard)
class Clipboard(BaseClipboard):

    #---------------------------------------------------------------------------
    #  'data' property methods:
    #---------------------------------------------------------------------------

    def _get_has_data(self):
        return self.has_object_data or self.has_text_data or self.has_file_data

    #---------------------------------------------------------------------------
    #  'object_data' property methods:
    #---------------------------------------------------------------------------

    def _get_object_data(self):
        obj = None
        mime_data = cb.mimeData()
        if mime_data.hasFormat(PYTHON_TYPE):
            serialized_data = BytesIO(mime_data.data(PYTHON_TYPE).data())
            klass = load(serialized_data)
            obj = load(serialized_data)
        return obj

    def _set_object_data(self, data):
        mime_data = QtCore.QMimeData()
        serialized_data = dumps(data.__class__) + dumps(data)
        mime_data.setData(PYTHON_TYPE, QtCore.QByteArray(serialized_data))
        cb.setMimeData(mime_data)

    def _get_has_object_data(self):
        return cb.mimeData().hasFormat(PYTHON_TYPE)

    def _get_object_type(self):
        result = ''
        mime_data = cb.mimeData()
        if mime_data.hasFormat(PYTHON_TYPE):
            try:
                # We may not be able to load the required class:
                result = loads(mime_data.data(PYTHON_TYPE).data())
            except:
                pass
        return result

    #---------------------------------------------------------------------------
    #  'text_data' property methods:
    #---------------------------------------------------------------------------

    def _get_text_data(self):
        return cb.text()

    def _set_text_data(self, data):
        cb.setText(data)

    def _get_has_text_data(self):
        return cb.mimeData().hasText()

    #---------------------------------------------------------------------------
    #  'file_data' property methods:
    #---------------------------------------------------------------------------

    def _get_file_data(self):
        mime_data = cb.mimeData()
        if mime_data.hasUrls():
            return [url.toString() for url in mime_data.urls()]
        else:
            return []

    def _set_file_data(self, data):
        if isinstance(data, str):
            data = [data]
        mime_data = QtCore.QMimeData()
        mime_data.setUrls([QtCore.QUrl(path) for path in data])
        cb.setMimeData(mime_data)

    def _get_has_file_data (self):
        return cb.mimeData().hasUrls()





from traits.testing.unittest_tools import unittest

from ..heading_text import HeadingText
from ..split_application_window import SplitApplicationWindow
from ..toolkit import toolkit_object

GuiTestAssistant = toolkit_object('util.gui_test_assistant:GuiTestAssistant')
no_gui_test_assistant = (GuiTestAssistant.__name__ == 'Unimplemented')


@unittest.skipIf(no_gui_test_assistant, 'No GuiTestAssistant')
class TestSplitApplicationWindow(unittest.TestCase, GuiTestAssistant):

    def setUp(self):
        GuiTestAssistant.setUp(self)
        self.window = SplitApplicationWindow()

    def tearDown(self):
        if self.window.control is not None:
            with self.delete_widget(self.window.control):
                self.window.destroy()
        del self.window
        GuiTestAssistant.tearDown(self)

    def test_destroy(self):
        # test that destroy works even when no control
        self.window.destroy()

    def test_open_close(self):
        # test that opening and closing works as expected
        with self.assertTraitChanges(self.window, 'opening', count=1):
            with self.assertTraitChanges(self.window, 'opened', count=1):
                self.window.open()
        self.event_loop()
        with self.assertTraitChanges(self.window, 'closing', count=1):
            with self.assertTraitChanges(self.window, 'closed', count=1):
                self.window.close()
        self.event_loop()

    def test_horizontal_split(self):
        # test that horizontal split works
        self.window.direction = 'horizontal'
        with self.assertTraitChanges(self.window, 'opening', count=1):
            with self.assertTraitChanges(self.window, 'opened', count=1):
                self.window.open()
        self.event_loop()
        with self.assertTraitChanges(self.window, 'closing', count=1):
            with self.assertTraitChanges(self.window, 'closed', count=1):
                self.window.close()
        self.event_loop()

    def test_contents(self):
        # test that contents works
        self.window.lhs = HeadingText
        self.window.rhs = HeadingText
        with self.assertTraitChanges(self.window, 'opening', count=1):
            with self.assertTraitChanges(self.window, 'opened', count=1):
                self.window.open()
        self.event_loop()
        with self.assertTraitChanges(self.window, 'closing', count=1):
            with self.assertTraitChanges(self.window, 'closed', count=1):
                self.window.close()
        self.event_loop()

    def test_ratio(self):
        # test that ratio split works
        self.window.ratio = 0.25
        with self.assertTraitChanges(self.window, 'opening', count=1):
            with self.assertTraitChanges(self.window, 'opened', count=1):
                self.window.open()
        self.event_loop()
        with self.assertTraitChanges(self.window, 'closing', count=1):
            with self.assertTraitChanges(self.window, 'closed', count=1):
                self.window.close()
        self.event_loop()



"""
Module defining a simple Python shell task.

This task provides a view with a simple Python shell.  This shouldn't be
confused with a more full-featured shell, such as those provided by IPython.
"""

# Std lib imports
import logging

# Enthought library imports.
from traits.api import Str, List, Dict, Instance
from pyface.api import PythonShell, FileDialog, OK
from pyface.tasks.api import Task, TaskPane
from pyface.tasks.action.api import SMenuBar, SMenu, TaskAction

# set up logging
logger = logging.getLogger()


class PythonShellPane(TaskPane):
    """ A Tasks Pane containing a Pyface PythonShell
    """
    id = 'pyface.tasks.contrib.python_shell.pane'
    name = 'Python Shell'
    
    editor = Instance(PythonShell)
    
    bindings = List(Dict)
    commands = List(Str)
    
    def create(self, parent):
        """ Create the python shell task pane
        
        This wraps the standard pyface PythonShell
        """
        logger.debug('PythonShellPane: creating python shell pane')
        self.editor = PythonShell(parent)
        self.control = self.editor.control
        
        # bind namespace
        logger.debug('PythonShellPane: binding variables')
        for binding in self.bindings:
            for name, value in list(binding.items()):
                self.editor.bind(name, value)
        
        # execute commands
        logger.debug('PythonShellPane: executing startup commands')
        for command in self.commands:
            self.editor.execute_command(command)
               
        logger.debug('PythonShellPane: created')
    
    def destroy(self):
        """ Destroy the python shell task pane
        """
        logger.debug('PythonShellPane: destroying python shell pane')
        self.editor.destroy()
        self.control = self.editor = None
        logger.debug('PythonShellPane: destroyed')


class PythonShellTask(Task):
    """
    A task which provides a simple Python Shell to the user.
    """
    
    # Task Interface
    
    id = 'pyface.tasks.contrib.python_shell'
    name = 'Python Shell'
    
    # The list of bindings for the shell
    bindings = List(Dict)
    
    # The list of commands to run on shell startup
    commands = List(Str)
    
    # the IPythonShell instance that we are interacting with
    pane = Instance(PythonShellPane)
    
    # Task Interface

    menu_bar = SMenuBar(SMenu(TaskAction(name='Open...', method='open',
                                         accelerator='Ctrl+O'),
                              id='File', name='&File'),
                        SMenu(id='View', name='&View'))
    
    def create_central_pane(self):
        """ Create a view pane with a Python shell
        """
        logger.debug("Creating Python shell pane in central pane")
        self.pane = PythonShellPane(bindings=self.bindings, commands=self.commands)
        return self.pane

    # PythonShellTask API

    def open(self):
        """ Shows a dialog to open a file.
        """
        logger.debug('PythonShellTask: opening file')
        dialog = FileDialog(parent=self.window.control, wildcard='*.py')
        if dialog.open() == OK:
            self._open_file(dialog.path)

    # Private API
    
    def _open_file(self, path):
        """ Execute the selected file in the editor's interpreter
        """
        logger.debug('PythonShellTask: executing file "%s"' % path)
        self.pane.editor.execute_file(path)


#------------------------------------------------------------------------------
# Copyright (c) 2010, Enthought Inc
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.

#
# Author: Enthought Inc
# Description: <Enthought pyface code editor>
#------------------------------------------------------------------------------

from pyface.qt import QtGui

from pygments.lexer import RegexLexer, _TokenType, Text, Error
from pygments.lexers import CLexer, CppLexer, PythonLexer, get_lexer_by_name
from pygments.styles.default import DefaultStyle
from pygments.token import Comment


def get_tokens_unprocessed(self, text, stack=('root',)):
    """ Split ``text`` into (tokentype, text) pairs.

        Monkeypatched to store the final stack on the object itself.

        The `text` parameter that gets passed is only the current line, so to
        highlight things like multiline strings correctly, we need to retrieve
        the state from the previous line (this is done in PygmentsHighlighter,
        below), and use it to continue processing the current line.
    """
    pos = 0
    tokendefs = self._tokens
    if hasattr(self, '_saved_state_stack'):
        statestack = list(self._saved_state_stack)
    else:
        statestack = list(stack)
    statetokens = tokendefs[statestack[-1]]
    while 1:
        for rexmatch, action, new_state in statetokens:
            m = rexmatch(text, pos)
            if m:
                if action is not None:
                    if type(action) is _TokenType:
                        yield pos, action, m.group()
                    else:
                        for item in action(self, m):
                            yield item
                pos = m.end()
                if new_state is not None:
                    # state transition
                    if isinstance(new_state, tuple):
                        for state in new_state:
                            if state == '#pop':
                                statestack.pop()
                            elif state == '#push':
                                statestack.append(statestack[-1])
                            else:
                                statestack.append(state)
                    elif isinstance(new_state, int):
                        # pop
                        del statestack[new_state:]
                    elif new_state == '#push':
                        statestack.append(statestack[-1])
                    else:
                        assert False, "wrong state def: %r" % new_state
                    statetokens = tokendefs[statestack[-1]]
                break
        else:
            try:
                if text[pos] == '\n':
                    # at EOL, reset state to "root"
                    pos += 1
                    statestack = ['root']
                    statetokens = tokendefs['root']
                    yield pos, Text, '\n'
                    continue
                yield pos, Error, text[pos]
                pos += 1
            except IndexError:
                break
    self._saved_state_stack  = list(statestack)

# Monkeypatch!
RegexLexer.get_tokens_unprocessed = get_tokens_unprocessed


# Even with the above monkey patch to store state, multiline comments do not
# work since they are stateless (Pygments uses a single multiline regex for
# these comments, but Qt lexes by line). So we need to add a state for comments
# to the C and C++ lexers. This means that nested multiline comments will appear
# to be valid C/C++, but this is better than the alternative for now.

def replace_pattern(tokens, new_pattern):
    """ Given a RegexLexer token dictionary 'tokens', replace all patterns that
        match the token specified in 'new_pattern' with 'new_pattern'.
    """
    for state in list(tokens.values()):
        for index, pattern in enumerate(state):
            if isinstance(pattern, tuple) and pattern[1] == new_pattern[1]:
                state[index] = new_pattern

# More monkeypatching!
comment_start = (r'/\*', Comment.Multiline, 'comment')
comment_state = [ (r'[^*/]', Comment.Multiline),
                  (r'/\*', Comment.Multiline, '#push'),
                  (r'\*/', Comment.Multiline, '#pop'),
                  (r'[*/]', Comment.Multiline) ]
replace_pattern(CLexer.tokens, comment_start)
replace_pattern(CppLexer.tokens, comment_start)
CLexer.tokens['comment'] = comment_state
CppLexer.tokens['comment'] = comment_state


class BlockUserData(QtGui.QTextBlockUserData):
    """ Storage for the user data associated with each line.
    """

    syntax_stack = ('root',)

    def __init__(self, **kwds):
        QtGui.QTextBlockUserData.__init__(self)
        for key, value in kwds.items():
            setattr(self, key, value)

    def __repr__(self):
        attrs = ['syntax_stack']
        kwds = ', '.join([ '%s=%r' % (attr, getattr(self, attr))
                           for attr in attrs ])
        return 'BlockUserData(%s)' % kwds


class PygmentsHighlighter(QtGui.QSyntaxHighlighter):
    """ Syntax highlighter that uses Pygments for parsing. """

    def __init__(self, parent, lexer=None):
        super(PygmentsHighlighter, self).__init__(parent)

        try:
            self._lexer = get_lexer_by_name(lexer)
        except:
            self._lexer = PythonLexer()

        self._style = DefaultStyle
        # Caches for formats and brushes.
        self._brushes = {}
        self._formats = {}

    def highlightBlock(self, qstring):
        """ Highlight a block of text.
        """
        qstring = str(qstring)
        prev_data = self.previous_block_data()

        if prev_data is not None:
            self._lexer._saved_state_stack = prev_data.syntax_stack
        elif hasattr(self._lexer, '_saved_state_stack'):
            del self._lexer._saved_state_stack

        index = 0
        # Lex the text using Pygments
        for token, text in self._lexer.get_tokens(qstring):
            l = len(text)
            format = self._get_format(token)
            if format is not None:
                self.setFormat(index, l, format)
            index += l

        if hasattr(self._lexer, '_saved_state_stack'):
            data = BlockUserData(syntax_stack=self._lexer._saved_state_stack)
            self.currentBlock().setUserData(data)

            # there is a bug in pyside and it will crash unless we
            # hold on to the reference a little longer
            data = self.currentBlock().userData()

            # Clean up for the next go-round.
            del self._lexer._saved_state_stack

    def previous_block_data(self):
        """ Convenience method for returning the previous block's user data.
        """
        return self.currentBlock().previous().userData()

    def _get_format(self, token):
        """ Returns a QTextCharFormat for token or None.
        """
        if token in self._formats:
            return self._formats[token]
        result = None
        for key, value in list(self._style.style_for_token(token).items()):
            if value:
                if result is None:
                    result = QtGui.QTextCharFormat()
                if key == 'color':
                    result.setForeground(self._get_brush(value))
                elif key == 'bgcolor':
                    result.setBackground(self._get_brush(value))
                elif key == 'bold':
                    result.setFontWeight(QtGui.QFont.Bold)
                elif key == 'italic':
                    result.setFontItalic(True)
                elif key == 'underline':
                    result.setUnderlineStyle(
                        QtGui.QTextCharFormat.SingleUnderline)
                elif key == 'sans':
                    result.setFontStyleHint(QtGui.QFont.SansSerif)
                elif key == 'roman':
                    result.setFontStyleHint(QtGui.QFont.Times)
                elif key == 'mono':
                    result.setFontStyleHint(QtGui.QFont.TypeWriter)
                elif key == 'border':
                    # Borders are normally used for errors. We can't do a border
                    # so instead we do a wavy underline
                    result.setUnderlineStyle(
                        QtGui.QTextCharFormat.WaveUnderline)
                    result.setUnderlineColor(self._get_color(value))
        self._formats[token] = result
        return result

    def _get_brush(self, color):
        """ Returns a brush for the color.
        """
        result = self._brushes.get(color)
        if result is None:
            qcolor = self._get_color(color)
            result = QtGui.QBrush(qcolor)
            self._brushes[color] = result

        return result

    def _get_color(self, color):
        qcolor = QtGui.QColor()
        qcolor.setRgb(int(color[:2],base=16),
                      int(color[2:4], base=16),
                      int(color[4:6], base=16))
        return qcolor




#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Enthought, Inc.
#
#------------------------------------------------------------------------------

""" Enthought pyface package component
"""

# Major package imports.
import wx

# Enthought library imports.
from traits.api import Bool, provides, Unicode

# Local imports.
from pyface.i_directory_dialog import IDirectoryDialog, MDirectoryDialog
from .dialog import Dialog


@provides(IDirectoryDialog)
class DirectoryDialog(MDirectoryDialog, Dialog):
    """ The toolkit specific implementation of a DirectoryDialog.  See the
    IDirectoryDialog interface for the API documentation.
    """


    #### 'IDirectoryDialog' interface #########################################

    default_path = Unicode

    message = Unicode

    new_directory = Bool(True)

    path = Unicode

    ###########################################################################
    # Protected 'IDialog' interface.
    ###########################################################################

    def _create_contents(self, parent):
        # In wx this is a canned dialog.
        pass

    ###########################################################################
    # 'IWindow' interface.
    ###########################################################################

    def close(self):
        # Get the path of the chosen directory.
        self.path = str(self.control.GetPath())

        # Let the window close as normal.
        super(DirectoryDialog, self).close()

    ###########################################################################
    # Protected 'IWidget' interface.
    ###########################################################################

    def _create_control(self, parent):
        # The default style.
        style = wx.OPEN

        # Create the wx style depending on which buttons are required etc.
        if self.new_directory:
            style = style | wx.DD_NEW_DIR_BUTTON

        if self.message:
            message = self.message
        else:
            message = "Choose a directory"

        # Create the actual dialog.
        return wx.DirDialog(parent, message=message,
                defaultPath=self.default_path, style=style)

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described in the PyQt GPL exception also apply

#
# Author: Riverbank Computing Limited
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------


# Major package imports.
from pyface.qt import QtCore, QtGui

# Enthought library imports.
from traits.api import Int, provides, Unicode

# Local imports.
from pyface.i_heading_text import IHeadingText, MHeadingText
from .widget import Widget


@provides(IHeadingText)
class HeadingText(MHeadingText, Widget):
    """ The toolkit specific implementation of a HeadingText.  See the
    IHeadingText interface for the API documentation.
    """


    #### 'IHeadingText' interface #############################################

    level = Int(1)

    text = Unicode('Default')

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, parent, **traits):
        """ Creates the panel. """

        # Base class constructor.
        super(HeadingText, self).__init__(**traits)

        # Create the toolkit-specific control that represents the widget.
        self._create_control(parent)

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_control(self, parent):
        """ Create the toolkit-specific control that represents the widget. """

        self.control = QtGui.QLabel(parent)
        self._set_text(self.text)

        self.control.setFrameShape(QtGui.QFrame.StyledPanel)
        self.control.setFrameShadow(QtGui.QFrame.Raised)
        self.control.setSizePolicy(QtGui.QSizePolicy.Preferred,
                                   QtGui.QSizePolicy.Fixed)

    def _set_text(self, text):
        """ Set the text on the toolkit specific widget. """

        # Bold the text. Qt supports a limited subset of HTML for rich text.
        text = "<b>" + text + "</b>"

        self.control.setText(text)

    #### Trait event handlers #################################################

    def _text_changed(self, new):
        """ Called when the text is changed. """

        if self.control is not None:
            self._set_text(new)

#### EOF ######################################################################




from pickle import dumps, load, loads, PickleError
import warnings
import io
import sys

from pyface.qt import QtCore


#-------------------------------------------------------------------------------
#  'PyMimeData' class:
#-------------------------------------------------------------------------------

if sys.version_info[0] < 3:
    def str2bytes(s):
        return s
else:
    def str2bytes(s):
        return bytes(s,'ascii')


class PyMimeData(QtCore.QMimeData):
    """ The PyMimeData wraps a Python instance as MIME data.
    """
    # The MIME type for instances.
    MIME_TYPE = 'application/x-ets-qt4-instance'
    NOPICKLE_MIME_TYPE = 'application/x-ets-qt4-instance-no-pickle'

    def __init__(self, data=None, pickle=True):
        """ Initialise the instance.
        """
        QtCore.QMimeData.__init__(self)

        # Keep a local reference to be returned if possible.
        self._local_instance = data

        if pickle:
            if data is not None:
                # We may not be able to pickle the data.
                try:
                    pdata = dumps(data)
                    # This format (as opposed to using a single sequence) allows
                    # the type to be extracted without unpickling the data.
                    self.setData(self.MIME_TYPE, dumps(data.__class__) + pdata)
                except (PickleError, TypeError, AttributeError):
                    # if pickle fails, still try to create a draggable
                    warnings.warn(("Could not pickle dragged object %s, " +
                            "using %s mimetype instead") % (repr(data),
                            self.NOPICKLE_MIME_TYPE), RuntimeWarning)
                    self.setData(self.NOPICKLE_MIME_TYPE, str2bytes(str(id(data))))

        else:
            self.setData(self.NOPICKLE_MIME_TYPE, str2bytes(str(id(data))))

    @classmethod
    def coerce(cls, md):
        """ Wrap a QMimeData or a python object to a PyMimeData.
        """
        # See if the data is already of the right type.  If it is then we know
        # we are in the same process.
        if isinstance(md, cls):
            return md

        if isinstance(md, PyMimeData):
            # if it is a PyMimeData, migrate all its data, subclasses should
            # override this method if it doesn't do thgs correctly for them
            data = md.instance()
            nmd = cls()
            nmd._local_instance = data
            for format in md.formats():
                nmd.setData(format, md.data(format))
        elif isinstance(md, QtCore.QMimeData):
            # if it is a QMimeData, migrate all its data
            nmd = cls()
            for format in md.formats():
                nmd.setData(format, md.data(format))
        else:
            # by default, try to pickle the coerced object
            pickle = True

            # See if the data is a list, if so check for any items which are
            # themselves of the right type.  If so, extract the instance and
            # track whether we should pickle.
            # XXX lists should suffice for now, but may want other containers
            if isinstance(md, list):
                pickle = not any(item.hasFormat(cls.NOPICKLE_MIME_TYPE)
                        for item in md if isinstance(item, QtCore.QMimeData))
                md = [item.instance() if isinstance(item, PyMimeData) else item
                        for item in md]

            # Arbitrary python object, wrap it into PyMimeData
            nmd = cls(md, pickle)

        return nmd

    def instance(self):
        """ Return the instance.
        """
        if self._local_instance is not None:
            return self._local_instance

        if not self.hasFormat(self.MIME_TYPE):
            # We have no pickled python data defined.
            return None

        stream = io.BytesIO(self.data(self.MIME_TYPE).data())

        try:
            # Skip the type.
            load(stream)

            # Recreate the instance.
            return load(stream)
        except PickleError:
            pass

        return None

    def instanceType(self):
        """ Return the type of the instance.
        """
        if self._local_instance is not None:
            return self._local_instance.__class__

        try:
            if self.hasFormat(self.MIME_TYPE):
                return loads(self.data(self.MIME_TYPE).data())
        except PickleError:
            pass

        return None

    def localPaths(self):
        """ The list of local paths from url list, if any.
        """
        ret = []
        for url in self.urls():
            if url.scheme() == 'file':
                ret.append(url.toLocalFile())
        return ret



#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described in the PyQt GPL exception also apply

#
# Author: Riverbank Computing Limited
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------


# Major package imports.
from pyface.qt import QtCore


class Timer(QtCore.QTimer):
    """Simple subclass of QTimer that allows the user to have a function called
    periodically.  Some code assumes that this is a sub-class of wx.Timer, so
    we add a few wrapper methods to pretend it is.

    Any exceptions raised in the callable are caught.  If
    `StopIteration` is raised the timer stops.  If other exceptions are
    encountered the timer is stopped and the exception re-raised.
    """

    def __init__(self, millisecs, callable, *args, **kw_args):
        """ Initialize instance to invoke the given `callable` with given
        arguments and keyword args after every `millisecs` (milliseconds).
        """
        QtCore.QTimer.__init__(self)

        self.callable = callable
        self.args = args
        self.kw_args = kw_args

        self.timeout.connect(self.Notify)

        self._is_active = True
        self.start(millisecs)

    def Notify(self):
        """ Call the given callable.  Exceptions raised in the callable are
        caught.  If `StopIteration` is raised the timer stops.  If other
        exceptions are encountered the timer is stopped and the exception
        re-raised.  Note that the name of this method is part of the API
        because some code expects this to be a wx.Timer sub-class.
        """
        try:
            self.callable(*self.args, **self.kw_args)
        except StopIteration:
            self.stop()
        except:
            self.stop()
            raise

    def Start(self, millisecs=None):
        """ Emulate wx.Timer.
        """
        self._is_active = True

        if millisecs is None:
            self.start()
        else:
            self.start(millisecs)

    def Stop(self):
        """ Emulate wx.Timer.
        """
        self._is_active = False
        self.stop()

    def IsRunning(self):
        """ Emulate wx.Timer.
        """
        return self._is_active





from traits.testing.unittest_tools import unittest

from ..clipboard import clipboard


class TestObject(object):
    def __init__(self, **kwargs):
        for key, value in list(kwargs.items()):
            setattr(self, key, value)

    def __eq__(self, other):
        if isinstance(other, TestObject):
            return all(getattr(other, key) == value
                       for key, value in list(self.__dict__.items()))


class TestClipboard(unittest.TestCase):

    def setUp(self):
        self.clipboard = clipboard

    def test_set_text_data(self):
        self.clipboard.data = 'test'
        self.assertTrue(self.clipboard.has_data)
        self.assertEqual(self.clipboard.data_type, 'str')
        self.assertEqual(self.clipboard.data, 'test')
        self.assertTrue(self.clipboard.has_text_data)
        self.assertEqual(self.clipboard.text_data, 'test')
        self.assertFalse(self.clipboard.has_file_data)
        self.assertFalse(self.clipboard.has_object_data)

    def test_set_text_data_unicode(self):
        self.clipboard.data = 'test'
        self.assertTrue(self.clipboard.has_data)
        self.assertEqual(self.clipboard.data_type, 'str')
        self.assertEqual(self.clipboard.data, 'test')
        self.assertTrue(self.clipboard.has_text_data)
        self.assertEqual(self.clipboard.text_data, 'test')
        self.assertFalse(self.clipboard.has_file_data)
        self.assertFalse(self.clipboard.has_object_data)

    @unittest.skip('backends not consistent')
    def test_set_file_data(self):
        self.clipboard.data = ['file:///images']
        self.assertTrue(self.clipboard.has_data)
        self.assertEqual(self.clipboard.data_type, 'file')
        self.assertEqual(self.clipboard.data, ['/images'])
        self.assertTrue(self.clipboard.has_file_data)
        self.assertEqual(self.clipboard.file_data, ['/images'])
        self.assertFalse(self.clipboard.has_text_data)
        self.assertFalse(self.clipboard.has_object_data)

    def test_set_object_data(self):
        data = TestObject(foo='bar', baz=1)
        self.clipboard.data = data
        self.assertTrue(self.clipboard.has_data)
        self.assertEqual(self.clipboard.data_type, TestObject)
        self.assertEqual(self.clipboard.data, data)
        self.assertTrue(self.clipboard.has_object_data)
        self.assertEqual(self.clipboard.object_type, TestObject)
        self.assertEqual(self.clipboard.object_data, data)
        self.assertFalse(self.clipboard.has_text_data)
        self.assertFalse(self.clipboard.has_file_data)





import platform

from traits.testing.unittest_tools import unittest

from ..constant import CANCEL, NO, OK, YES
from ..toolkit import toolkit_object
from ..window import Window


is_qt = toolkit_object.toolkit == 'qt4'
if is_qt:
    from pyface.qt import qt_api

GuiTestAssistant = toolkit_object('util.gui_test_assistant:GuiTestAssistant')
no_gui_test_assistant = (GuiTestAssistant.__name__ == 'Unimplemented')

ModalDialogTester = toolkit_object('util.modal_dialog_tester:ModalDialogTester')
no_modal_dialog_tester = (ModalDialogTester.__name__ == 'Unimplemented')

is_pyqt5 = (is_qt and qt_api == 'pyqt5')
is_pyqt4_linux = (is_qt and qt_api == 'pyqt' and platform.system() == 'Linux')


@unittest.skipIf(no_gui_test_assistant, 'No GuiTestAssistant')
class TestWindow(unittest.TestCase, GuiTestAssistant):

    def setUp(self):
        GuiTestAssistant.setUp(self)
        self.window = Window()

    def tearDown(self):
        if self.window.control is not None:
            with self.delete_widget(self.window.control):
                self.window.destroy()
        self.window = None
        GuiTestAssistant.tearDown(self)

    def test_destroy(self):
        # test that destroy works even when no control
        self.window.destroy()
        self.event_loop()

    def test_open_close(self):
        # test that opening and closing works as expected
        with self.assertTraitChanges(self.window, 'opening', count=1):
            with self.assertTraitChanges(self.window, 'opened', count=1):
                self.window.open()
        self.event_loop()
        with self.assertTraitChanges(self.window, 'closing', count=1):
            with self.assertTraitChanges(self.window, 'closed', count=1):
                self.window.close()
        self.event_loop()

    def test_show(self):
        # test that showing works as expected
        self.window._create()
        self.window.show(True)
        self.event_loop()
        self.window.show(False)
        self.event_loop()
        self.window.destroy()
        self.event_loop()

    def test_activate(self):
        # test that activation works as expected
        self.window.open()
        self.event_loop()
        self.window.activate()
        self.event_loop()
        self.window.close()
        self.event_loop()

    def test_position(self):
        # test that default position works as expected
        self.window.position = (100, 100)
        self.window.open()
        self.event_loop()
        self.window.close()
        self.event_loop()

    def test_reposition(self):
        # test that changing position works as expected
        self.window.open()
        self.event_loop()
        self.window.position = (100, 100)
        self.event_loop()
        self.window.close()
        self.event_loop()

    def test_size(self):
        # test that default size works as expected
        self.window.size = (100, 100)
        self.window.open()
        self.event_loop()
        self.window.close()
        self.event_loop()

    def test_resize(self):
        # test that changing size works as expected
        self.window.open()
        self.event_loop()
        self.window.size = (100, 100)
        self.event_loop()
        self.window.close()
        self.event_loop()

    def test_title(self):
        # test that default title works as expected
        self.window.title = "Test Title"
        self.window.open()
        self.event_loop()
        self.window.close()
        self.event_loop()

    def test_retitle(self):
        # test that changing title works as expected
        self.window.open()
        self.event_loop()
        self.window.title = "Test Title"
        self.event_loop()
        self.window.close()
        self.event_loop()

    def test_show_event(self):
        self.window.open()
        self.window.visible = False
        self.event_loop()

        with self.assertTraitChanges(self.window, 'visible', count=1):
            self.window.control.show()
            self.event_loop()

        self.assertTrue(self.window.visible)

    def test_hide_event(self):
        self.window.open()
        self.event_loop()

        with self.assertTraitChanges(self.window, 'visible', count=1):
            self.window.control.hide()
            self.event_loop()

        self.assertFalse(self.window.visible)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_confirm_reject(self):
        # test that cancel works as expected
        tester = ModalDialogTester(
            lambda: self.window.confirm("message", cancel=True))
        tester.open_and_run(when_opened=lambda x: x.close(accept=False))
        self.assertEqual(tester.result, CANCEL)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    @unittest.skipIf(is_pyqt5, "Confirmation dialog click tests don't work on pyqt5.")  # noqa
    @unittest.skipIf(is_pyqt4_linux, "Confirmation dialog click tests don't work reliably on linux.  Issue #282.")  # noqa
    def test_confirm_yes(self):
        # test that yes works as expected
        tester = ModalDialogTester(lambda: self.window.confirm("message"))
        tester.open_and_wait(when_opened=lambda x: x.click_button(YES))
        self.event_loop()
        self.assertEqual(tester.result, YES)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    @unittest.skipIf(is_pyqt5, "Confirmation dialog click tests don't work on pyqt5.")  # noqa
    @unittest.skipIf(is_pyqt4_linux, "Confirmation dialog click tests don't work reliably on linux.  Issue #282.")  # noqa
    def test_confirm_no(self):
        # test that no works as expected
        tester = ModalDialogTester(lambda: self.window.confirm("message"))
        tester.open_and_wait(when_opened=lambda x: x.click_button(NO))
        self.event_loop()
        self.assertEqual(tester.result, NO)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    @unittest.skipIf(is_pyqt5, "Confirmation dialog click tests don't work on pyqt5.")  # noqa
    @unittest.skipIf(is_pyqt4_linux, "Confirmation dialog click tests don't work reliably on linux.  Issue #282.")  # noqa
    def test_confirm_cancel(self):
        # test that cncel works as expected
        tester = ModalDialogTester(
            lambda: self.window.confirm("message", cancel=True))
        tester.open_and_wait(when_opened=lambda x: x.click_button(CANCEL))
        self.event_loop()
        self.assertEqual(tester.result, CANCEL)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_information_accept(self):
        self._check_message_dialog_accept(self.window.information)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    @unittest.skipIf(is_pyqt5, "Message dialog click tests don't work on pyqt5.")  # noqa
    @unittest.skipIf(is_pyqt4_linux, "Message dialog click tests don't work reliably on linux.  Issue #282.")  # noqa
    def test_information_ok(self):
        self._check_message_dialog_ok(self.window.information)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_warning_accept(self):
        self._check_message_dialog_accept(self.window.warning)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    @unittest.skipIf(is_pyqt5, "Message dialog click tests don't work on pyqt5.")  # noqa
    @unittest.skipIf(is_pyqt4_linux, "Message dialog click tests don't work reliably on linux.  Issue #282.")  # noqa
    def test_warning_ok(self):
        self._check_message_dialog_ok(self.window.warning)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_error_accept(self):
        self._check_message_dialog_accept(self.window.error)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    @unittest.skipIf(is_pyqt5, "Message dialog click tests don't work on pyqt5.")  # noqa
    @unittest.skipIf(is_pyqt4_linux, "Message dialog click tests don't work reliably on linux.  Issue #282.")  # noqa
    def test_error_ok(self):
        self._check_message_dialog_ok(self.window.error)

    def _check_message_dialog_ok(self, method):
        tester = self._setup_tester(method)
        tester.open_and_wait(when_opened=lambda x: x.click_button(OK))
        self.assertIsNone(tester.result)

    def _check_message_dialog_accept(self, method):
        tester = self._setup_tester(method)
        tester.open_and_run(when_opened=lambda x: x.close(accept=True))
        self.assertIsNone(tester.result)

    def _setup_tester(self, method):
        kwargs = {
            'title': 'Title',
            'detail': 'Detail',
            'informative': 'Informative'
        }
        tester = ModalDialogTester(lambda: method("message", **kwargs))
        return tester



#------------------------------------------------------------------------------
# Copyright (c) 2010, Enthought Inc
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.

#
# Author: Enthought Inc
# Description: <Enthought pyface code editor>
#------------------------------------------------------------------------------

import weakref

from pyface.qt import QtGui, QtCore

from .find_widget import FindWidget

class ReplaceWidget(FindWidget):

    def __init__(self, parent):
        super(FindWidget, self).__init__(parent)
        self.adv_code_widget = weakref.ref(parent)

        self.button_size = self.fontMetrics().width('Replace All') + 20

        form_layout = QtGui.QFormLayout()
        form_layout.addRow('Fin&d', self._create_find_control())
        form_layout.addRow('Rep&lace', self._create_replace_control())

        layout = QtGui.QHBoxLayout()
        layout.addLayout(form_layout)

        close_button = QtGui.QPushButton('Close')
        layout.addWidget(close_button, 1, QtCore.Qt.AlignRight)
        close_button.clicked.connect(self.hide)

        self.setLayout(layout)

    def _create_replace_control(self):
        control = QtGui.QWidget(self)

        self.replace_edit = QtGui.QLineEdit()
        self.replace_button = QtGui.QPushButton('&Replace')
        self.replace_button.setFixedWidth(self.button_size)
        self.replace_all_button = QtGui.QPushButton('Replace &All')
        self.replace_all_button.setFixedWidth(self.button_size)

        layout = QtGui.QHBoxLayout()
        layout.addWidget(self.replace_edit)
        layout.addWidget(self.replace_button)
        layout.addWidget(self.replace_all_button)
        layout.addStretch(2)
        layout.setContentsMargins(0, 0, 0, 0)

        control.setLayout(layout)
        return control



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The implementation of a dialog that displays a message. """




# Convenience functions.
def information(parent, message, title='Information',
                detail='', informative=''):
    """ Convenience method to show an information message dialog.

    Parameters
    ----------
    parent : toolkit control or None
        The toolkit control that should be the parent of the dialog.
    message : str
        The text of the message to display.
    title : str
        The text of the dialog title.
    detail : str
        Further details about the message (displayed when the user clicks
        "Show details").
    informative : str
        Explanatory text to display along with the message.

    """
    dialog = MessageDialog(
        parent=parent, message=message, title=title,
        severity='information', detail=detail, informative=informative
    )
    dialog.open()


def warning(parent, message, title='Warning', detail='', informative=''):
    """ Convenience function to show a warning message dialog.

    Parameters
    ----------
    parent : toolkit control or None
        The toolkit control that should be the parent of the dialog.
    message : str
        The text of the message to display.
    title : str
        The text of the dialog title.
    detail : str
        Further details about the message (displayed when the user clicks
        "Show details").
    informative : str
        Explanatory text to display along with the message.

    """
    dialog = MessageDialog(
        parent=parent, message=message, title=title,
        severity='warning', detail=detail, informative=informative
    )
    dialog.open()


def error(parent, message, title='Error', detail='', informative=''):
    """ Convenience function to show an error message dialog.

    Parameters
    ----------
    parent : toolkit control or None
        The toolkit control that should be the parent of the dialog.
    message : str
        The text of the message to display.
    title : str
        The text of the dialog title.
    detail : str
        Further details about the message (displayed when the user clicks
        "Show details").
    informative : str
        Explanatory text to display along with the message.

    """
    dialog = MessageDialog(
        parent=parent, message=message, title=title, severity='error',
        detail=detail, informative=informative
    )
    dialog.open()


# Import the toolkit specific version.
from .toolkit import toolkit_object
MessageDialog = toolkit_object('message_dialog:MessageDialog')

#### EOF ######################################################################



""" An action that sets an attribute. """


# Enthought library imports.
from traits.api import Any, Str

# Local imports.
from .workbench_action import WorkbenchAction


class SetattrAction(WorkbenchAction):
    """ An action that sets an attribute. """

    #### 'SetattrAction' interface ############################################

    # The object that we set the attribute on.
    obj = Any

    # The name of the attribute that we set.
    attribute_name = Str

    # The value that we set the attribute to.
    value = Any

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Performs the action. """

        setattr(self.obj, self.attribute_name, self.value)

        return

#### EOF ######################################################################
